// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.collections;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import xl2.collections.node.ListNode;

  /**
  * A singly linked list partial-implementation of the java.util.List 
  * interface.
  * <p>Most functionality of the list interface does not require a doubly linked
  * list.  However, USE a doubly linked list if the following methods are
  * needed:<ul>
  * <li>SList.lastIndexOf(Object) - must scan the entire list.
  * <li>SList.listIterator().hasPrevious() - throws 
  *     UnsupportedOperationException.
  * <li>SList.listIterator().previous() - throws 
  *     UnsupportedOperationException.
  * <li>SList.listIterator().previousIndex() - throws 
  *     UnsupportedOperationException.
  * </ul>
  * <p>SubList(int,int) does not return a list backed by the original 
  * like java.util.LinkedList(). This returns a new list to eliminate any 
  * ambiguous behavior.
  * @author Aaron Hansen
  * @version 1.0.0
  */
public class SList implements java.util.List {

  // Public Fields
  // -------------

  public static transient int UNLIMITED_CAPACITY = -1;


  // Protected Fields
  // ----------------

    /**
    * Max elements in the list. &lt;0 meaning no limit, and the default.
    */
  protected int CapacityInt = UNLIMITED_CAPACITY;

    /**
    * First/beginning/top... node.
    */
  protected ListNode HeadNode = null;

    /**
    * Incremented every time the list is modified.
    */
  protected transient int ModInt = 0;

    /**
    * Last/ending/bottom... node.
    */
  protected ListNode TailNode = null;

    /**
    * The number of elements in the list.
    */
  protected int SizeInt = 0;


  // Constructors
  // ------------

  public SList() {}

  public SList(Collection collection) {
    addAll(collection);
    }


  // Public Methods
  // --------------

  public boolean add(Object element) {
    return (addAfter(element,TailNode) != null);
    }

  public void add(int index, Object element) {
    if (index == 0)
      addAfter(element,null);
    else
      addAfter(element,getNode(--index));
    }

    /**
    * Adds an element after the first occurrence of the specified element.
    * @param addition Whom to add.
    * @param preceding If null, the element is inserted at the head
    *        of the list.
    * @return False if list at full capacity, otherwise true.
    */
  public boolean addAfter(Object addition, Object preceding) {
    return (addAfter(addition,getNode(preceding)) != null);
    }

  public boolean addAll(Collection collection) {
    if (collection.size() == 0)
      return false;
    Iterator iterator = collection.iterator();
    while (iterator.hasNext())
      addAfter(iterator.next(),TailNode);
    return true;
    }

  public boolean addAll(int index, Collection collection) {
    if (collection.size() == 0)
      return false;
    if (index == SizeInt) //append
      return addAll(collection);
    SList list = new SList(collection);
    if (index == 0) { //prepend
      if (list.TailNode != null) {
        list.TailNode.NextNode = HeadNode;
        HeadNode = list.HeadNode;
        }
      }
    else { //insert
      ListNode node = getNode(--index); //get node before insertion point
      if (list.TailNode != null) {
        list.TailNode.NextNode = node.NextNode;
        node.NextNode = list.HeadNode;
        }
      }
    SizeInt += list.SizeInt;
    ModInt++;  //this is how java.util.LinkedList handles bulk additions.
    return true;
    }

    /**
    * Prepends element to the list.
    * @return False If at full capacity.
    * @see #getCapacity()
    */
  public boolean addFirst(Object element) {
    return (addAfter(element,null) != null);
    }

    /**
    * Releases the list but lets the garbage collector do the work.
    */
  public void clear() {
    HeadNode = null;
    TailNode = null;
    SizeInt = 0;
    ModInt++;
    }

  public boolean contains(Object element) {
    return (getNode(element) != null);
    }

  public boolean containsAll(Collection collection) {
    Iterator iterator = collection.iterator();
    while (iterator.hasNext()) {
      if (getNode(iterator.next()) == null)
        return false;
      }
    return true;
    }

  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (obj == null)
      return false;
    if (!(obj instanceof List))
      return false;
    Collection coll = (Collection) obj;
    if (coll.size() != this.size())
      return false;
    Iterator mine = iterator();
    Iterator it = coll.iterator();
    Object myObj;
    Object itsObj;
    while (mine.hasNext()) {
      myObj = mine.next();
      itsObj = it.next();
      if (myObj != itsObj) {
        if ((myObj == null) || (itsObj == null) || !myObj.equals(itsObj))
          return false;
        }
      }
    return true;
    }

  public Object get(int index) {
    try {
      return getNode(index).DataObject;
      } catch (Exception x) {}
    return null;
    }

    /**
    * Max allowable elements in the structure (&lt;0 is unlimited).
    * @return &lt;0 is no limit, and is the default.
    * @see #UNLIMITED_CAPACITY
    */
  public int getCapacity() {
    return CapacityInt;
    }

  public Object getFirst() {
    if (HeadNode == null)
      return null;
    return HeadNode.DataObject;
    }

  public Object getLast() {
    if (TailNode == null)
      return null;
    return TailNode.DataObject;
    }

    /**
    * Uses the hash algorithm specifed by java.util.List - 
    * which is inefficient, System.identityHashCode(Object) may be preferable.
    */
  public int hashCode() {
    int hash = 1;
    Iterator iterator = iterator();
    Object obj;
    while (iterator.hasNext()) {
      obj = iterator.next();
      hash = 31 * hash + (obj==null ? 0 : obj.hashCode());
      }
    return hash;
    }

  public int indexOf(Object element) {
    int index = 0;
    ListNode node = HeadNode;
    if (element == null) {
      while (node != null) {
        if (node.DataObject == null)
          return index;
        node = node.NextNode;
        index++;
        }
      }
    else {
      while (node != null) {
        if (element.equals(node.DataObject))
          return index;
        node = node.NextNode;
        index++;
        }
      }
    return -1;
    }

  public boolean isEmpty() {
    return (SizeInt == 0);
    }

  public Iterator iterator() {
    return new SListIterator(this,0,-1);
    }

    /**
    * A doubly linked list is recommended if frequent use of this method is
    * required.
    */
  public int lastIndexOf(Object element) {
    int last = -1;
    int index = 0;
    ListNode node = HeadNode;
    if (element == null) {
      while (node != null) {
        if (node.DataObject == null)
          last = index;
        node = node.NextNode;
        index++;
        }
      }
    else {
      while (node != null) {
        if (element.equals(node.DataObject))
          last = index;
        node = node.NextNode;
        index++;
        }
      }
    return last;
    }

    /**
    * Returns a ListIterator which throws UnsupportedOperationException from
    * the hasPrevious(), previous() and previousIndex() methods.
    */
  public ListIterator listIterator() {
    return new SListIterator(this,0,-1);
    }

    /**
    * Returns a ListIterator which throws UnsupportedOperationException from
    * the hasPrevious(), previous() and previousIndex() methods.
    */
  public ListIterator listIterator(int fromIndex) {
    return new SListIterator(this,fromIndex,-1);
    }

    /**
    * Debug, prints contents of list to System.out
    */
  public void print() {
    print(System.out);
    }

    /**
    * Debug, prints contents of list to the stream.
    */
  public void print(PrintStream out) {
    ListNode node = HeadNode;
    while (node != null) {
      out.println(String.valueOf(node.DataObject));
      node = node.NextNode;
      }
    out.flush();
    }

    /**
    * Debug, prints contents of list to the stream.
    */
  public void print(PrintWriter out) {
    PrintWriter writer = new PrintWriter(out);
    ListNode node = HeadNode;
    while (node != null) {
      writer.println(String.valueOf(node.DataObject));
      node = node.NextNode;
      }
    writer.flush();
    }

  public Object remove(int index) {
    ListNode node = getNode(index);
    removeNode(node);
    return node.DataObject;
    }

  public boolean remove(Object element) {
    if (removeNodeOf(element) == null)
      return false;
    return true;
    }

  public boolean removeAll(Collection c) {
    int startmod = ModInt;
    Iterator iterator = iterator();
    Object obj = null;
    while (iterator.hasNext()) {
      obj = iterator.next();
      if (c.contains(obj))
        iterator.remove();
      }
    return (startmod != ModInt);
    }

    /**
    * Removes and returns the first element of the list.
    * @return The first element, or null.
    */
  public Object removeFirst() {
    if (HeadNode == null)
      return null;
    Object obj = HeadNode.DataObject;
    if (removeNode(HeadNode))
      return obj;
    else
      return null;
    }

  public boolean retainAll(Collection c) {
    int startmod = ModInt;
    Iterator iterator = iterator();
    Object obj = null;
    while (iterator.hasNext()) {
      obj = iterator.next();
      if (!c.contains(obj))
        iterator.remove();
      }
    return (startmod != ModInt);
    }

  public Object set(int index, Object element) {
    ListNode node = getNode(index);
    Object old = node.DataObject;
    node.DataObject = element;
    ModInt++;
    return old;
    }

    /**
    * The maximum allowable elements of the list.  Setting the capacity less
    * than the current size will not modify the elements of the list -
    * it just behaves as if at full capacity.
    */
  public void setCapacity(int capacity) {
    CapacityInt = capacity;
    }

  public int size() {
    return SizeInt;
    }

    /**
    * Returns a new list (not backed by this) representing a portion of 
    * this list between from (included) and to (excluded).
    * @param from First element of the sublist.
    * @param to First excluded element of the sublist.
    * @return A new list.
    */
  public List subList(int from, int to) {
    Iterator iterator = new SListIterator(this,from,to);
    SList sub = new SList();
    while (iterator.hasNext())
      sub.add(iterator.next());
    return sub;
    }

  public Object[] toArray() {
    Iterator iterator = iterator();
    Object[] ret = new Object[SizeInt];
    int i = -1;
    while (iterator.hasNext())
      ret[++i] = iterator.next();
    return ret;
    }

  public Object[] toArray(Object[] tryToUse) {
    if (tryToUse.length < SizeInt)
      tryToUse = (Object[]) java.lang.reflect.Array.newInstance(
                            tryToUse.getClass().getComponentType(), SizeInt);
    int i = 0;
    for (ListNode curr = HeadNode; curr != null; curr = curr.NextNode)
      tryToUse[i++] = curr.DataObject;
    if (tryToUse.length > SizeInt)
      tryToUse[SizeInt] = null;
    return tryToUse;
    }


  // Protected Methods
  // -----------------

    /**
    * @param predecessor Null inserts the addition at the head of the list.
    * @return The node containing the addition, or null if the list is full
    *         to capacity.
    */
  protected ListNode addAfter(Object addition, ListNode predecessor) {
    if ((SizeInt >= CapacityInt) && (CapacityInt >= 0))
      return null;
    ListNode node = new ListNode(addition);
    if (predecessor == null) { //new start of list
      if (HeadNode == null) {
        HeadNode = node;
        TailNode = node;
        }
      else {
        node.NextNode = HeadNode;
        HeadNode = node;
        }
      }
    else {
      node.NextNode = predecessor.NextNode;
      predecessor.NextNode = node;
      if (TailNode == predecessor)
        TailNode = node;
      }
    SizeInt++;
    ModInt++;
    return node;
    }

  protected ListNode getHead() {
    return HeadNode;
    }

  protected ListNode getNode(int index) {
    if ((index < 0) || (index >= SizeInt))
      throw new IndexOutOfBoundsException(String.valueOf(index));
    int i = 0;
    ListNode node = HeadNode;
    while (i < index) {
      node = node.NextNode;
      i++;
      }
    return node;
    }

    /**
    * Returns the first node encountered whose element equals the parameter.
    */
  protected ListNode getNode(Object element) {
    ListNode node = HeadNode;
    if (element == null) {
      while (node != null) {
        if (node.DataObject == null)
          return node;
        node = node.NextNode;
        }
      }
    else {
      while (node != null) {
        if (element.equals(node.DataObject))
          return node;
        node = node.NextNode;
        }
      }
    return null;
    }

  protected ListNode getTail() {
    return TailNode;
    }

  protected boolean removeNode(ListNode node) {
    ListNode curr = HeadNode;
    ListNode prev = null;
    while ((curr != node) & (curr != null)) {
      prev = curr;
      curr = curr.NextNode;
      }
    if (curr == null)
      return false;
    if (curr == HeadNode)
      HeadNode = HeadNode.NextNode;
    if (curr == TailNode)
      TailNode = prev;
    if (prev != null)
      prev.NextNode = curr.NextNode;
    SizeInt--;
    ModInt++;
    return true;
    }

  protected ListNode removeNodeOf(Object element) {
    ListNode prev = null;
    ListNode node = HeadNode;
    if (element == null) {
      while (node != null) {
        if (node.DataObject == null) {
          if (node == HeadNode)
            HeadNode = node.NextNode;
          if (node == TailNode)
            TailNode = prev;
          if (prev != null)
            prev.NextNode = node.NextNode;
          SizeInt--;
          ModInt++;
          return node;
          }
        prev = node;
        node = node.NextNode;
        }
      }
    else {
      while (node != null) {
        if (element.equals(node.DataObject)) {
          if (node == HeadNode)
            HeadNode = node.NextNode;
          if (node == TailNode)
            TailNode = prev;
          if (prev != null)
            prev.NextNode = node.NextNode;
          SizeInt--;
          ModInt++;
          return node;
          }
        prev = node;
        node = node.NextNode;
        }
      }
    return null;
    }


  }//class
