// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.collections;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.AbstractMap;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedMap;

  /**
  * A skip list implementation of the SortedMap interface.  It is
  * backed by a SkipSet.
  * <p>
  * This collection is not synchronized.  Note that it is safe for any
  * number of threads to query the collection while a thread is mutating it.  
  * It is not safe for multiple threads to concurrently modify the collection.
  * <p>
  * This is a high performance collection but it can not provide guaranteed
  * execution time because of its random nature.  At the lowest level
  * a skip list is a linked list, thus iterators are lightning fast.
  * <p>
  * A minimal number of nodes are affected by addition or removal.  This is 
  * advantage in persistence as the size of the transaction is minimized.
  * <p>
  * For those familiar with skip lists, this implementation automagically 
  * maintains levels and probabilities.  Levels are added, but never 
  * removed unless the set is cleared.
  * @author Aaron Hansen
  * @version 1.0.0
  */
public class SkipMap extends AbstractMap implements Externalizable, SortedMap {

  // Protected Fields
  // ----------------

  protected EntrySet TheEntrySet = null;


  // Constructors
  // ------------

  public SkipMap() { 
    this(null,null);
    }

  public SkipMap(Comparator comparator) { 
    this(null,comparator);
    }

  public SkipMap(Map map) { 
    this(map,null);
    }

  public SkipMap(Map map, Comparator comparator) { 
    TheEntrySet = new EntrySet(new SkipSet(comparator));
    if (map != null) {
      Iterator iterator = map.entrySet().iterator();
      Map.Entry node;
      SkipSet set = TheEntrySet.getSet();
      while (iterator.hasNext()) {
        node = (Map.Entry) iterator.next();
        set.put(node.getKey(),node.getValue());
        }
      }
    }

  public SkipMap(SortedMap map) { 
    this(map,map.comparator());
    }

  protected SkipMap(EntrySet set) {
    TheEntrySet = set;
    }


  // Public Methods
  // --------------

  public Comparator comparator() {
    return TheEntrySet.comparator();
    }

  public boolean containsKey(Object key) {
    return TheEntrySet.contains(key);
    }

  public Set entrySet() {
    return TheEntrySet;
    }

  public Object firstKey() {
    return TheEntrySet.first();
    }

  public Object get(Object key) {
    return TheEntrySet.get(key);
    }

  public SortedMap headMap(Object to) {
    return subMap(null,to);
    }

  public Object lastKey() {
    return TheEntrySet.last();
    }

  public Object put(Object key, Object value) {
    return TheEntrySet.put(key,value);
    }

  public void readExternal(ObjectInput in) 
    throws ClassNotFoundException, IOException {
    TheEntrySet = (EntrySet) in.readObject();
    }

  public Object remove(Object key) {
    return TheEntrySet.removeKey(key);
    }

  public SortedMap subMap(Object from, Object to) {
    return new SkipMap((EntrySet)TheEntrySet.subSet(from,to));
    }

  public SortedMap tailMap(Object from) {
    return subMap(from,null);
    }

  public Collection values() {
    return new ValueCollection(this);
    }

  public void writeExternal(ObjectOutput out) throws IOException {
    out.writeObject(TheEntrySet);
    }


  }//SkipMap
