// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.collections; 

import java.io.*;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.SortedSet;
import xl2.collections.node.SkipNode;
import xl2.collections.node.SkipMapNode;

  /**
  * A skip list implementation of the SortedSet interface.  
  * <p>
  * This collection is not synchronized.  Note that it is safe for any
  * number of threads to query the collection while a thread is mutating it.  
  * It is not safe for multiple threads to concurrently modify the collection.
  * <p>
  * This is a high performance collection but it can not provide guaranteed
  * execution time because of its random nature.  At the lowest level
  * a skip list is a linked list, thus iterators are lightning fast.
  * <p>
  * A minimal number of nodes are affected by addition or removal.  This is 
  * advantage in persistence as the size of the transaction is minimized.
  * <p>
  * For those familiar with skip lists, this implementation automagically 
  * maintains levels and probabilities.  Levels are added, but never 
  * removed unless the set is cleared.
  * @author Aaron Hansen
  * @version 1.0.0
  */
public class SkipSet extends AbstractSet implements Externalizable, SortedSet {

  // Protected Fields
  // ----------------

  protected SkipNode[] Levels = {null};

  protected transient long MaxNodesForLevel = 0;

  protected transient int ModInt = 0;

  protected transient double ProbabilityInc = 0;

  protected long SizeLong = 0;

  protected Comparator TheComparator = null;


  // Constructors
  // ------------

  public SkipSet() {
    this(null,null);
    }

  public SkipSet(Collection collection) {
    this(collection,null);
    }

  public SkipSet(Collection collection, Comparator comparator) {
    TheComparator = comparator;
    addLevel(); addLevel();
    addLevel(); addLevel();
    if (collection != null) {
      Iterator iterator = collection.iterator();
      while (iterator.hasNext())
        add(iterator.next());
      }
    }

  public SkipSet(Comparator comparator) {
    this(null,comparator);
    }

  public SkipSet(SortedSet set) {
    this(set,set.comparator());
    }


  // Public Methods
  // --------------

  public boolean add(Object obj) {
    if (obj == null)
      return false;
    SkipNode[] prev = predecessors(obj);
    if ( (prev[0] != null)
      && (prev[0].Next[0] != null) 
      && (compare(prev[0].Next[0].Key,obj) == 0) )
      return false;
    append(prev,new SkipNode(levels(),obj));
    return true;
    }

    /**
    * Returns the element that would be after the parameter - 
    * regardless if the parameter is contained or not.
    */
  public Object after(Object obj) {
    SkipNode node = successor(obj);
    if (node == null)
      return null;
    return node.Key;
    }

    /**
    * Returns the element that would be before the parameter - 
    * regardless if the parameter is contained or not.
    */
  public Object before(Object obj) {
    SkipNode node = predecessor(obj);
    if (node == null)
      return null;
    return node.Key;
    }

  public void clear() {
    Levels = new SkipNode[1];
    Levels[0] = null;
    addLevel(); addLevel();
    addLevel(); addLevel();
    SizeLong = 0;
    ModInt++;
    }

  public Comparator comparator() {
    return TheComparator;
    }

  public boolean contains(Object obj) {
    if (obj == null)
      return false;
    return nodeOf(obj) != null;
    }

    /**
    * First element of the set.
    */
  public Object first() {
    if (Levels[0] == null)
      return null;
    else
      return Levels[0].getValue();
    }

    /**
    * Returns an iterator of values of the set.
    */
  public Iterator iterator() {
    return new SkipValueIterator(this,Levels[0],null);
    }

  public SortedSet headSet(Object to) {
    return new SubSkipSet(this,null,to);
    }

    /**
    * Last element of the set.
    */
  public Object last() {
    SkipNode last = lastNode();
    if (last != null)
      return last.getValue();
    return null;
    }

    /**
    * True representation of the set's size.
    */
  public long longSize() {
    return SizeLong;
    }

    /**
    * Prints the set contents.  The format is: [level] value.toString().
    */
  public void print() {
    SkipNode current = Levels[0];
    System.out.println("[level] key.toString()");
    while (current != null) {
      System.out.print('[');
      System.out.print(current.Next.length);
      System.out.print(']');
      System.out.print(' ');
      System.out.println(current.Key.toString());
      current = current.Next[0];
      }
    }

  public void readExternal(ObjectInput in) 
    throws ClassNotFoundException, IOException {
    TheComparator = (Comparator) in.readObject();
    for (long l = in.readLong(); --l >= 0; )
      add(in.readObject());
    }

  public boolean remove(Object obj) {
    return (removeKey(obj) != null);
    }

    /**
    * This will populate a SkipSet with n objects and then show the 
    * distribution of objects at each level.
    */
  public static void test(int n) {
    SkipSet set = new SkipSet();
    long start = System.currentTimeMillis();
    for (int i = n; --i >= 0; )
      set.add(new Integer(i));
    System.out.println(
      "Time to populate set: " + (System.currentTimeMillis() - start));
    SkipNode current = set.Levels[0];
    int[] levels = new int[set.Levels.length];
    for (int i = levels.length; --i >= 0; )
      levels[i] = 0;
    while (current != null) {
      levels[current.Next.length - 1]++;
      current = current.Next[0];
      }
    for (int i = levels.length; --i >= 0; )
      System.out.println("Level " + i + " entries = " + levels[i]);
    }

  public int size() {
    if (SizeLong > Integer.MAX_VALUE)
      return Integer.MAX_VALUE;
    else
      return (int) SizeLong;
    }

  public SortedSet subSet(Object from, Object to) {
    return new SubSkipSet(this,from,to);
    }

  public SortedSet tailSet(Object from) {
    return new SubSkipSet(this,from,null);
    }

  public void writeExternal(ObjectOutput out) throws IOException {
    out.writeObject(TheComparator);
    Iterator iterator = iterator();
    out.writeLong(SizeLong);
    while (iterator.hasNext())
      out.writeObject(iterator.next());
    }


  // Protected Methods
  // -----------------

  protected void append(SkipNode[] prev, SkipNode node) {
    int level = 0;
    int levels = node.Next.length;
    while ((level < levels) && (prev[level] != null)) {
      node.Next[level] = prev[level].Next[level];
      prev[level].Next[level] = node;
      level++;
      }
    while (level < levels) {
      node.Next[level] = Levels[level];
      Levels[level] = node;
      level++;
      }
    if (++SizeLong > MaxNodesForLevel)
      addLevel();
    ModInt++;
    }

  protected void addLevel() {
    SkipNode[] nodes = new SkipNode[Levels.length + 1];
    System.arraycopy(Levels,0,nodes,0,Levels.length);
    nodes[Levels.length] = null;
    Levels = nodes;
    ProbabilityInc = .7d / (double) Levels.length;
    MaxNodesForLevel = (long) Math.pow(2,Levels.length);
    }

    /**
    * Returns the first node whose key is equal or greater, or 
    * null for a null key.
    */
  protected SkipNode ceiling(Object key) {
    if (key == null)
      return null;
    SkipNode current = predecessor(key);
    if (current == null)
      return Levels[0];
    else
      return current.Next[0];
    }

  protected int compare(Object obj1, Object obj2) {
    if (TheComparator == null)
      return obj1.hashCode() - obj2.hashCode();
    return TheComparator.compare(obj1,obj2);
    }

    /**
    * Returns the first node whose key is equal or greater, or
    * the first node for a null key.
    */
  protected SkipNode floor(Object obj) {
    if (obj == null)
      return Levels[0];
    SkipNode current = predecessor(obj);
    if (current == null)
      return Levels[0];
    else
      return current.Next[0];
    }

    /**
    * Returns this, used for getting the true set of a subset.
    * @see SubSkipSet#getSet()
    */
  protected SkipSet getSet() {
    return this;
    }

  protected SkipNode lastNode() {
    int level = Levels.length;
    SkipNode current = Levels[--level];
    while ((current == null) && (--level >= 0))
      current = Levels[level];
    if (current == null)
      return null;
    while (level >= 0) {
      if (current.Next[level] == null) 
        level--;
      else 
        current = current.Next[level];
      }
    return current;
    }

    /**
    * Randomly generates a number of levels (not an index).
    */
  protected int levels() {
    int levels = 1;
    double prob = 0.3d;
    while ((levels < Levels.length) && (Math.random() < prob)) {
      levels++;  //level, will be the length of the new node's Levels array.
      prob += ProbabilityInc;
      }
    return levels;
    }

  protected SkipNode nodeOf(Object obj) {
    SkipNode current = predecessor(obj);
    if (current == null) {
      if ((Levels[0] != null) && (compare(Levels[0].Key,obj) == 0))
        return Levels[0];
      else
        return null;
      }
    else if ((current.Next[0] != null) 
      && (compare(current.Next[0].Key,obj) == 0))
      return current.Next[0];
    return null;
    }

    /**
    * Returns the last node whose key is strictly less than the specified key.
    * The predescessor to null is the last item in the list.
    */
  protected SkipNode predecessor(Object obj) {
    if (obj == null)
      return lastNode();
    int level = Levels.length;
    SkipNode current = null;
    while ((current == null) && (--level >= 0)) { //find starting level
      if ( (Levels[level] != null) 
        && (compare(Levels[level].Key,obj) < 0) ) {
        current = Levels[level];
        }
      }
    if (current == null)
      return null;
    while (level >= 0) {  //now search
      while ((current.Next[level] != null) 
        &&  (compare(current.Next[level].Key,obj) < 0) ) {
        current = current.Next[level]; //go forward in level
        }
      level--;
      }
    return current;
    }
 
    /**
    * Returns all previous nodes, at all levels.  A null slot implies
    * the node is the first at the level.
    */
  protected SkipNode[] predecessors(Object obj) {
    int level = Levels.length;
    SkipNode[] prev = new SkipNode[level];
    SkipNode current = null;
    while ((current == null) && (--level >= 0)) { //find starting level
      if ((Levels[level] != null) && (compare(Levels[level].Key,obj) < 0))
        current = Levels[level];
      else
        prev[level] = current;
      }
    while (level >= 0) {  //now search for prev's
      while ((current.Next[level] != null) 
        &&  (compare(current.Next[level].Key,obj) < 0))
        current = current.Next[level]; //go forward in level
      prev[level] = current;
      level--;
      }
    return prev;
    }

    /**
    * For SkipMap wrapper.
    */
  protected Object put(Object key, Object value) {
    if (key == null)
      throw new NullPointerException();
    SkipNode[] prev = predecessors(key);
    if ( (prev[0] != null)
      && (prev[0].Next[0] != null) 
      && (compare(prev[0].Next[0].Key,key) == 0) )
      return prev[0].Next[0].setValue(value);
    append(prev, new SkipMapNode(levels(),key,value));
    return null;
    }

    /**
    * For SkipMap, returns value.
    */
  protected Object removeKey(Object obj) {
    if (obj == null)
      return null;
    SkipNode[] prev = predecessors(obj);
    SkipNode current = prev[0];
    if (current == null)
      current = Levels[0];
    else
      current = current.Next[0];
    if ((current == null) || (compare(current.Key,obj) != 0))
      return null;
    int level = 0;
    int levels = current.Next.length;
    while ((level < levels) && (prev[level] != null)) {
      prev[level].Next[level] = current.Next[level];
      level++;
      }
    while (level < levels) { 
      Levels[level] = current.Next[level];
      level++;
      }
    SizeLong--;
    ModInt++;
    return current.getValue();
    }

    /**
    * Returns the first node whose key is greater than the specified key.
    * The successor to null is the first item in the list.
    */
  protected SkipNode successor(Object obj) {
    if (obj == null)
      return Levels[0];
    int level = Levels.length;
    SkipNode current = null;
    while ((current == null) && (--level >= 0)) { //find starting level
      if ( (Levels[level] != null) 
        && (compare(Levels[level].Key,obj) <= 0) ) {
        current = Levels[level];
        }
      }
    if (current == null) {
      if ((Levels[0] != null) && (compare(Levels[0].Key,obj) > 0))
        return Levels[0];
      return null;
      }
    while (level >= 0) {  //now search
      while ((current.Next[level] != null) 
        &&  (compare(current.Next[level].Key,obj) <= 0) ) {
        current = current.Next[level]; //go forward in level
        }
      level--;
      }
    return current.Next[0];
    }
 

  }//SkipSet
