// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.collections;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.AbstractSet; 
import java.util.Comparator;
import java.util.Iterator;
import java.util.SortedSet;
import xl2.collections.node.SkipNode;

  /**
  * Subset wrapper of a skip set.
  * @author Aaron Hansen
  * @see SkipSet
  * @version 1.0.0
  */
class SubSkipSet extends AbstractSet implements Externalizable, SortedSet {

  // Protected Fields
  // ----------------

  protected Object FromObject;

  protected int ModInt = 0;

  protected long SizeLong = -1;

  protected SkipSet TheSet = null;

  protected Object ToObject;


  // Constructors
  // ------------

  public SubSkipSet() {}

  public SubSkipSet(SkipSet set) {
    this(set,null,null);
    }

  public SubSkipSet(SkipSet set, Object first, Object to) {
    if (set == null)
      throw new NullPointerException("Null skip structure.");
    TheSet = set.getSet();
    FromObject = first;
    ToObject = to;
    }


  // Public Methods
  // --------------

  public boolean add(Object obj) {
    if (inRange(obj))
      return TheSet.add(obj);
    ModInt = TheSet.ModInt;
    SizeLong++;
    return false;
    }

  public void clear() {
    Iterator iterator = iterator();
    while (iterator.hasNext()) {
      iterator.next();
      iterator.remove();
      }
    ModInt = TheSet.ModInt;
    SizeLong = 0;
    }

  public Comparator comparator() {
    return TheSet.comparator();
    }

  public boolean contains(Object obj) {
    if (inRange(obj))
      return TheSet.contains(obj);
    return false;
    }

  public Object first() {
    SkipNode node = TheSet.floor(FromObject);
    if (node == null)
      return null;
    return node.getValue();
    }

  public SortedSet headSet(Object to) {
    if (!inRange(to))
      throw new IllegalArgumentException("Element out of range.");
    return new SubSkipSet(TheSet,FromObject,to);
    }

  public Iterator iterator() {
    return new SkipValueIterator(getSet(),floor(),ceiling());
    }

  public Object last() {
    SkipNode node = TheSet.predecessor(ToObject);
    if (node == null)
      return null;
    return node.getValue();
    }

  public long longSize() {
    return SizeLong;
    }

  public void readExternal(ObjectInput in) 
    throws ClassNotFoundException, IOException {
    TheSet = (SkipSet) in.readObject();
    FromObject = in.readObject();
    ToObject = in.readObject();
    ModInt = in.readInt();
    SizeLong = in.readLong();
    }

  public boolean remove(Object obj) {
    if (inRange(obj)) {
      if (TheSet.remove(obj)) {
        SizeLong--;
        ModInt = TheSet.ModInt;
        }
      }
    return false;
    }

  public int size() {
    if ((FromObject == null) && (ToObject == null))
      return TheSet.size();
    else if ((SizeLong < 0) || (ModInt != TheSet.ModInt)) {
      SizeLong = 0;
      ModInt = TheSet.ModInt;
      SkipNode current = TheSet.floor(FromObject);
      if (current == null)
        return 0;
      if (!inRange(current.getValue()))
        return 0;
      SkipNode to = TheSet.ceiling(ToObject);
      while (current.Next[0] != to) {
        current = current.Next[0];
        SizeLong++;
        }
      }
    if (SizeLong > Integer.MAX_VALUE)
      return Integer.MAX_VALUE;
    else
      return (int) SizeLong;
    }

  public SortedSet subSet(Object from, Object to) {
    if (!inRange(from))
      throw new IllegalArgumentException("From out of range.");
    if (!inRange(to))
      throw new IllegalArgumentException("To out of range.");
    if ((from != null) && (to != null) && (TheSet.compare(from,to) > 0))
      throw new IllegalArgumentException("From is greater than To.");
    return new SubSkipSet(TheSet,from,to);
    }

  public SortedSet tailSet(Object from) {
    if (!inRange(from))
      throw new IllegalArgumentException("Element out of range.");
    return new SubSkipSet(TheSet,from,ToObject);
    }

  public void writeExternal(ObjectOutput out) throws IOException {
    out.writeObject(TheSet);
    out.writeObject(FromObject);
    out.writeObject(ToObject);
    out.writeInt(ModInt);
    out.writeLong(SizeLong);
    }


  // Protected Methods
  // -----------------

  protected SkipNode ceiling() {
    return TheSet.ceiling(ToObject);
    }

  protected SkipNode ceiling(Object to) {
    return TheSet.ceiling(to);
    }

  protected SkipNode floor() {
    return TheSet.floor(FromObject);
    }

  protected SkipNode floor(Object from) {
    return TheSet.floor(from);
    }

  protected SkipSet getSet() {
    return TheSet;
    }

  protected boolean inRange(Object obj) {
    if ((FromObject != null) && (TheSet.compare(obj, FromObject) < 0))
      return false;
    if ((ToObject != null) && (TheSet.compare(obj, ToObject) > 0))
      return false;
    return true;
    }


  }//SubSkipSet
