// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.odb;

import java.io.*;
import java.util.*;
import xl2.collections.SkipSet;
import xl2.collections.node.SkipNode;

  /**
  * Maintains free space of a file.
  * Two sets are used to tract the free space.  One set is ordered by offset 
  * and the other by size of the data.  
  * Free space is acquired by its size and an offset is returned.  Any remaining
  * space from the chunck requested is recorded in the free list.
  * Adjoining space is alway combined, and if the file size can be shortened, 
  * it is.
  * Changes are not saved until commit() is called.  Abort will return the 
  * free list to its last successful commit() state.
  * @author Aaron Hansen
  * @version 1.0.0
  */
class FreeList implements Comparator {

  // Protected Fields
  // ----------------

  protected long CurrentLong = 0;

  protected RandomAccessFile ListRAFile = null;

  protected File ListFile = null;

  protected boolean ModifiedBoolean = false;

  protected RandomAccessFile SourceFile = null;

  protected long ResetLong = 0;

  protected LengthSet TheLengthSet = null;

  protected SkipSet OffSet = null;


  // Constructors
  // ------------

  public FreeList(RandomAccessFile source, File listFile) 
    throws IOException {
    SourceFile = source;
    ListFile = listFile;
    ListRAFile = new RandomAccessFile(listFile, "rw");
    init();
    }


  // Public Methods
  // --------------

    /**
    * Resets the state of the free list to the last successful commit()
    */
  public void abort() throws IOException {
    SourceFile.setLength(ResetLong);
    ModifiedBoolean = false;
    init();
    }

  protected void add(FreeEntry entry) {
    FreeEntry predecessor = (FreeEntry) OffSet.before(entry);
    if (predecessor != null) {
      if (predecessor.next() == entry.OffsetLong) {
        TheLengthSet.remove(predecessor);
        OffSet.remove(predecessor);
        entry.OffsetLong = predecessor.OffsetLong;
        entry.LengthInt += predecessor.LengthInt;
        }
      }
    FreeEntry successor = (FreeEntry) OffSet.after(entry);
    if (successor != null) {
      if (entry.next() == successor.OffsetLong) {
        TheLengthSet.remove(successor);
        OffSet.remove(successor);
        entry.LengthInt += successor.LengthInt;
        }
      }
    TheLengthSet.add(entry);
    OffSet.add(entry);
    ModifiedBoolean = true;
    }

  public void add(long offset, int length) {
    add(new FreeEntry(offset,length));
    }

  public void close() throws IOException {
    if (ListRAFile != null)
      ListRAFile.close();
    }

  public void commit() throws IOException {
    FreeEntry entry = (FreeEntry) OffSet.last();
    if ((entry != null) && (entry.next() == SourceFile.length())) {
      TheLengthSet.remove(entry);
      OffSet.remove(entry);
      CurrentLong = entry.OffsetLong;
      SourceFile.setLength(CurrentLong); 
      ModifiedBoolean = true;
      }
    if (ModifiedBoolean) {
      ListRAFile.setLength(0);
      DataOutputStream dos =
        new DataOutputStream(
          new BufferedOutputStream(
            new FileOutputStream(ListFile)));
      Iterator iterator = OffSet.iterator();
      dos.writeLong(OffSet.longSize());
      while (iterator.hasNext()) {
        entry = (FreeEntry) iterator.next();
        dos.writeLong(entry.OffsetLong);
        dos.writeInt(entry.LengthInt);
        }
      dos.flush();
      dos.close();
      }
    ResetLong = CurrentLong;
    ModifiedBoolean = false;
    }

  public int compare(Object obj1, Object obj2) {
    long l = ((FreeEntry)obj1).OffsetLong - ((FreeEntry)obj2).OffsetLong;
    if (l < 0)
      return -1;
    if (l > 0)
      return 1;
    return 0;
    }

    /**
    * Returns the entry at the lowest offset.
    */
  public FreeEntry first() {
    return (FreeEntry) OffSet.first();
    }

  public boolean isEmpty() {
    return (OffSet.size() == 0);
    }

    /**
    * Removes and returns an offset for the specified length.
    */
  public long next(int length) {
    FreeEntry entry = TheLengthSet.remove(length);
    long l = CurrentLong;
    if (entry == null)
      CurrentLong += length;
    else {
      ModifiedBoolean = true;
      l = entry.OffsetLong;
      OffSet.remove(entry);
      int diff = entry.LengthInt - length;
      if (diff > 0) {
        entry.OffsetLong += length;
        entry.LengthInt = diff;
        add(entry);
        }
      }
    return l;
    }

    /**
    * For debugging.
    */
  public void print() {
    System.out.println("[Free List --->");
    FreeEntry entry;
    Iterator iterator = OffSet.iterator();
    while (iterator.hasNext()) {
      entry = (FreeEntry) iterator.next();
      System.out.print(entry.OffsetLong);
      System.out.print(',');
      System.out.println(entry.LengthInt);
      }
    System.out.println("<--- Free List]");
    }

    /**
    * Removes and returns an entry at the lowest available offset.  May return
    * null;
    */
  public void remove(FreeEntry entry) {
    if (entry != null) {
      ModifiedBoolean = true;
      OffSet.remove(entry);
      TheLengthSet.remove(entry);
      }
    }


  // Protected Methods
  // -----------------

  protected void init() throws IOException {
    ResetLong = SourceFile.length();
    CurrentLong = ResetLong;
    if (TheLengthSet == null) {
      TheLengthSet = new LengthSet();
      OffSet = new SkipSet(this);
      }
    else {
      TheLengthSet.clear();
      OffSet.clear();
      }
    if (ListRAFile.length() == 0)
      return;
    DataInputStream dis = 
      new DataInputStream(
        new BufferedInputStream(
          new FileInputStream(ListFile)));
    FreeEntry entry;
    for (long l = dis.readLong(); --l >= 0; ) {
      entry = new FreeEntry(dis.readLong(),dis.readInt());
      TheLengthSet.add(entry);
      OffSet.add(entry);
      }
    dis.close();
    }


  }//FreeList


  /**
  * Stores FreeEntry's with duplicate lengths.
  */
class LengthSet extends SkipSet {

  // Constructors
  // ------------

  public LengthSet() {}

  public boolean add(Object obj) {
    if (obj == null)
      return false;
    SkipNode[] prev = predecessors(obj);
    if ( (prev[0] != null) 
      && (((FreeEntry)prev[0].Key).LengthInt == ((FreeEntry)obj).LengthInt))
      append(prev, new SkipNode(prev[0].Next.length,obj));
    else
      append(prev, new SkipNode(levels(),obj));
    return true;
    }

  public FreeEntry remove(int length) {
    //This is a trick, notice how compare considers all entries
    //with the same length unequal, unless the instances are the same.
    FreeEntry entry = (FreeEntry) after(new FreeEntry(0,--length));
    if (entry == null)
      return null;
    return (FreeEntry) removeKey(entry);
    }

  protected int compare(Object obj1, Object obj2) {
    int i = ((FreeEntry)obj1).LengthInt - ((FreeEntry)obj2).LengthInt;
    if ((i == 0) && (obj1 != obj2))
      return -1;
    return i;
    }

/*
  public boolean add(Object obj) {
    if (obj == null)
      return false;
    SkipNode[] prev = predecessors(obj);
    SkipNode node = prev[0];
    if (node == null)
      append(prev, new SkipNode(levels(),obj));
    else if (((FreeEntry)node.Key).LengthInt == ((FreeEntry)obj).LengthInt)
      append(prev, new SkipNode(node.Next.length,obj));
    else if (((node = node.Next[0]) != null) && 
      (((FreeEntry)node.Key).LengthInt == ((FreeEntry)obj).LengthInt))
      append(prev, new SkipNode(node.Next.length,obj));
    else
      append(prev, new SkipNode(levels(),obj));
    return true;
    }

  public FreeEntry remove(int length) {
    FreeEntry entry = (FreeEntry) after(new FreeEntry(Long.MAX_VALUE,--length));
    if (entry == null)
      return null;
    return (FreeEntry) removeKey(entry);
    }

  protected int compare(Object obj1, Object obj2) {
    int i = ((FreeEntry)obj1).LengthInt - ((FreeEntry)obj2).LengthInt;
    if ((i == 0) && (obj1 != obj2)) {
      long l = ((FreeEntry)obj1).OffsetLong - ((FreeEntry)obj2).OffsetLong;
      if (l > 0)
        return 1;
      if (l < 0)
        return -1;
      return 0;
      }
    return i;
    }
*/


  }//LengthSet
