// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.odb;

import xl2.odb.io.BAOS;
import xl2.odb.io.SeekableInputStream;
import xl2.odb.io.SeekableOutputStream;
import xl2.odb.util.ExecutionQueue;
import xl2.transaction.AbortFailedException;
import xl2.transaction.TransactionAbortedException;
import xl2.transaction.TransactionNotOpenException;
import java.io.*;
import java.util.*;

  /**
  * A network of serialized objects accessible from a single root object.
  * <p><b>Files</b><p>
  * An XL2 database is a directory containing of several files.<ul>
  * <li> freehdls.xl2 - Free space in the handles.xl2 file.
  * <li> freeobjs.xl2 - Free space in the objects.xl2 file.
  * <li> garbage.xl2 - Objects needing reclamation.
  * <li> handles.xl2 - Object identity that maps to data in objects.xl2.
  * <li> readlock.xl2.&lt;timeMillis&gt; - Lock file, there may be many.
  * <li> objects.xl2 - Serialized objects.
  * <li> writelock.xl2&lt;timeMillis&gt; - Lock file.
  * </ul>
  * <p><b>Usage</b><ul>
  * <li>Opening a database is independent of the XL2Database constructor.  
  * An XL2Database may be used to open multiple databases (not concurrently).
  * <li>XL2Databases and XL2References can only be accessed and mutated within 
  * the context of a transaction.  See method newTransaction().
  * </ul>
  * <p><b>Closing</b><ul>
  * <li>Always close a database. If the process of an open database 
  * terminates before closure, the database may become unrecoverable.  
  * <li>If the process of an open database crashes, the lock files(s) 
  * must be manually deleted.
  * <li>If JDK1.3 or higher is used, a shutdown hook will close all open
  * databases before the VM terminates.
  * </ul>
  * <p><b>Database Daemon</b><ul>
  * <li>Every database has a normal priority daemon thread which aborts 
  * expired transactions, collects garbage, and performs compaction.
  * <li>Garbage collection and compaction do not occur while there are other 
  * transactions.
  * </ul>
  * @author Aaron Hansen
  * @see XL2Transaction
  * @since Java 2
  * @version 1.1.0
  */
public class XL2Database {

  // Public Fields
  // -------------

    /**
    * Open mode.
    */
  public static final int NOT_OPEN = 0;
  
    /**
    * Open mode.
    */
  public static final int OPEN_READ_ONLY  = 1;
  
    /**
    * Open mode.
    */
  public static final int OPEN_READ_WRITE = 2;


  // Protected Fields
  // ----------------

  protected BAOS Baos = new BAOS(128);

    /**
    * Syncronizes database closing.
    */
  protected Object CloseLockObject = new Object();

    /**
    * Synchronizes and preserves order of threads closing a transaction.
    */
  protected ExecutionQueue CloseQueue = null;

    /**
    * Handle free list.
    */
  protected static final String FILE_FREE_HANDLE = "freehdls.xl2";

    /**
    * Object file free list.
    */
  protected static final String FILE_FREE_OBJECT = "freeobjs.xl2";
    
    /**
    * Garbage collected handles.
    */
  protected static final String FILE_GARBAGE     = "garbage.xl2";
    
    /**
    * Object handles.
    */
  protected static final String FILE_HANDLE      = "handles.xl2";

    /**
    * Lock file.
    */
  protected static final String FILE_READ_LOCK   = "readlock.xl2";

    /**
    * Serialized object file.
    */
  protected static final String FILE_OBJECT      = "objects.xl2";

    /**
    * Lock file.
    */
  protected static final String FILE_WRITE_LOCK  = "writelock.xl2";

    /**
    * Free handle file.
    */
  protected FreeStack HandleFreeStack = null;

  protected RandomAccessFile HandleFile = null ;

    /**
    * XL2Database locking file.
    */
  protected File LockFile = null ;

    /**
    * Synchronizes access to the object file.
    */
  protected Object LockObject = new Object();

    /**
    * The next available handle.
    */
  protected long NextHandleLong = 0L;

    /**
    * Object file reader.
    */
  protected SeekableInputStream ObjectReader = null;

    /**
    * Object file writer.
    */
  protected SeekableOutputStream ObjectWriter = null;

    /**
    * Free data file.
    */
  protected FreeList ObjFreeList = null;

  protected int OpenModeInt = NOT_OPEN;

    /**
    * XL2Database location.
    */
  protected String PathString = null;

    /**
    * Where to restore the next available handle in the event of abort.
    */
  protected long RestoreHandleLong = 0L;

  protected XL2Daemon TheDaemon = null;

    /**
    * The garbage file.
    */
  protected GarbageCollector TheGarbageCollector = null;

  protected HandleSet TheHandleSet = new HandleSet();

  protected static XL2ShutdownHook TheShutdownHook = null;

  static {
    try {
      XL2ShutdownHook hook = new XL2ShutdownHook();
      Runtime.getRuntime().addShutdownHook(hook);
      TheShutdownHook = hook;
      } catch (Throwable t) {};
    }

  protected Set XL2TransactionSet = new xl2.odb.util.HashSet();


  // Constructors
  // ------------

    /**
    * See open()
    */
  public XL2Database() {}
    

  // Public Methods
  // --------------

    /**
    * Aborts open transactions, evicts database contents from
    * memory, and unlocks the database.  Any exceptions are passed 
    * to log()
    */
  public void close() {
    synchronized (CloseLockObject) {
      if (OpenModeInt == NOT_OPEN)
        throw new DatabaseNotOpenException(this);
      CloseQueue.enqueue();
      TheDaemon.terminate();
      XL2Transaction[] trans = transactions();
      if (trans != null && (trans.length > 0)) {
        for (int i = trans.length; --i >= 0; ) {
          try { trans[i].abort(); } 
          catch (AbortFailedException x) { log(null,x); }
          catch (TransactionAbortedException x) {}
          catch (TransactionNotOpenException x) {}
          catch (Exception x) { log(null,x); }
          }
        }
      try { LockFile.delete(); } 
        catch (Exception x) { log(null,x); }
      try { ObjectReader.close(); } 
        catch (Exception x) { log(null,x); }
      try { ObjectWriter.close(); } 
        catch (Exception x) { log(null,x); }
      try { HandleFile.close(); } 
        catch (Exception x) { log(null,x); }
      try { HandleFreeStack.close(); } 
        catch (Exception x) { log(null,x); }
      try { ObjFreeList.close(); } 
        catch (Exception x) { log(null,x); }
      try { TheGarbageCollector.close(); } 
        catch (Exception x) { log(null,x); }
      if (TheShutdownHook != null)
        TheShutdownHook.removeDatabase(this);
      OpenModeInt = NOT_OPEN;
      }
    }

    /**
    * @deprecated Replaced by log(Throwable)
    * @see #log(Throwable)
    */
  public void closeException(Exception x) {
    log(null,x);
    }

    /**
    * Creates and returns a new XL2Database that will be open for read/write.
    * @param dbDir XL2Database directory.
    * @throws XL2Exception Wrapping any other exceptions.
    * @see #newDatabase(String)
    */
  public static synchronized XL2Database create(String dbDir) 
    throws XL2Exception{
    XL2Database db = new XL2Database();
    db.newDatabase(dbDir);
    db.open(dbDir,OPEN_READ_WRITE);
    return db;
    }

    /**
    * Callback for subclasses to manage serialization.
    * <ul>
    * <li>Called for every object being read from the database.
    * <li>Internally synchronized, subclasses needn't add any further 
    * synchronization.
    * <li>Streams will not be used concurrently with any other streams
    * from the same database instance - including those from
    * getObjectOutput(OutputStream)
    * </ul>
    * @see #getObjectOutput(OutputStream)
    * @throws XL2Exception Wrapping any internal exception such as IOExceptions.
    */
  public ObjectInput getObjectInput(InputStream in) throws XL2Exception {
    try {
      return new ObjectInputStream(in);
      }
    catch (Exception x) {
      log(null,x);
      throw new XL2Exception(x);
      }
    }

    /**
    * Callback for subclasses to manage serialization.
    * <ul>
    * <li>Called for every object being written to the database.
    * <li>Internally synchronized, subclasses needn't add any further 
    * synchronization.
    * <li>Streams will not be used concurrently with any other streams
    * from the same database instance - including those from
    * getObjectInput(InputStream)
    * <li>The stream will be closed by the database.
    * </ul>
    * @see #getObjectInput(InputStream)
    * @throws XL2Exception Wrapping any internal exception such as IOExceptions.
    */
  public ObjectOutput getObjectOutput(OutputStream out) throws XL2Exception {
    try {
      return new ObjectOutputStream(out);
      }
    catch (Exception x) {
      log(null,x);
      throw new XL2Exception(x);
      }
    }

    /**
    * NOT_OPEN, OPEN_READ_ONLY, or OPEN_READ_WRITE
    */
  public int getOpenMode() {
    return OpenModeInt;
    }

    /**
    * Returns the database entry object.
    */
  public synchronized Object getRoot() {
    if (OpenModeInt == NOT_OPEN)
      throw new DatabaseNotOpenException(this);
    return ((XL2Reference)get(0L)).get();
    }

    /**
    * Prints a timestamp, message and stack trace to "log.xl2" in 
    * the database directory.  If logging causes an exception, System.err 
    * is used.
    * @param message May be null.
    * @param exception May be null.
    */
  public void log(String message, Throwable throwable) {
    synchronized (PathString) {
      try {
        if (OpenModeInt == NOT_OPEN)
          throw new DatabaseNotOpenException(this);
        PrintWriter out =
          new PrintWriter(
            new FileWriter(
              new File(PathString,"log.xl2").getPath(),true));
        out.print("\n**[");
        out.print(new java.util.Date().toString());
        out.print("] ");
        if (message != null)
          out.println(message);
        else
          out.println();
        if (throwable != null)
          throwable.printStackTrace(out);
        out.close();
        }
      catch (Exception x) {
        System.err.print("\n**[");
        System.err.print(new java.util.Date().toString());
        System.err.println("] ");
        System.err.println(
          "Failure logging. Two stack traces follow.");
        System.err.println( 
          "First is what prevented the logging of the second.");
        throwable.printStackTrace(System.err);
        x.printStackTrace(System.err);
        }
      }
    }

    /**
    * Creates a database that will be closed upon completion of this method.
    * @param dbDir XL2Database directory.
    * @throws XL2Exception Wrapping any other exceptions.
    */
  public void newDatabase(String dbDir) throws XL2Exception {
    synchronized (XL2Database.class) {
      try {
        if (dbDir == null)
          throw new XL2Exception("No file name specified.");
        File dir = new File(dbDir);
        if (!dir.exists())
          dir.mkdir();
        File file;
        if (!dir.isDirectory())
          throw new XL2Exception("Not a directory: " + dbDir);
        if ((file = new File(dir,FILE_HANDLE)).exists())
          throw new XL2Exception("File already exists: " + file.getPath());
        file.createNewFile();
        if ((file = new File(dir,FILE_OBJECT)).exists())
          throw new XL2Exception("File already exists: " + file.getPath());
        file.createNewFile();
        if ((file = new File(dir,FILE_FREE_HANDLE)).exists())
          file.delete();
        file.createNewFile();
        if ((file = new File(dir,FILE_FREE_OBJECT)).exists())
          file.delete();
        file.createNewFile();
        if ((file = new File(dir,FILE_GARBAGE)).exists())
          file.delete();
        file.createNewFile();
        if ((file = new File(dir,FILE_WRITE_LOCK)).exists())
          file.delete();
        open(dbDir,OPEN_READ_WRITE);
        XL2Transaction trans = newTransaction();
        trans.begin();
        XL2Reference ref = new XL2Reference(0);
        trans.writeLock(ref);
        ReferentMgr.get(ref).setRefCount(1);
        trans.commit();
        close();
        }
      catch (XL2Exception x) {
        throw x;
        }
      catch (Exception x) {
        throw new XL2Exception(x);
        }
      }
    }

    /**
    * Required to access and mutate XL2References and opened 
    * database instances.
    * @throws DatabaseNotOpenException
    */
  public XL2Transaction newTransaction() {
    if (OpenModeInt == NOT_OPEN)
      throw new DatabaseNotOpenException(this);
    return new XL2Transaction(this);
    }

    /**
    * Opens the database with the given mode.
    * @param dbDir The database directory.
    * @param mode OPEN_READ_ONLY or OPEN_READ_WRITE
    * @throws DatabaseLockedException If a lock file prevents opening.
    * @throws XL2Exception Wrapping exceptions such as IOExceptions.
    */
  public void open(String dbDir, int mode) 
    throws DatabaseLockedException, XL2Exception {
    if ((mode != OPEN_READ_ONLY) && (mode != OPEN_READ_WRITE))
      throw new XL2Exception("Uknown open mode: " + mode);
    if (dbDir == null)
      throw new XL2Exception("No file name specified.");
    File dir = new File(dbDir);
    if (!dir.exists())
      throw new XL2Exception("XL2Database not found: " + dbDir);
    if (!dir.isDirectory())
      throw new XL2Exception("Not a directory: " + dbDir);
    PathString = dbDir;
    try {
      lock(dbDir,mode);
      File file = new File(dbDir,FILE_OBJECT);
      if (!file.exists())
        throw new XL2Exception("File does not exist" + file.getPath());
      ObjectReader = new SeekableInputStream(file,512);
      ObjectWriter = new SeekableOutputStream(file,512);
      file = new File(dbDir,FILE_HANDLE);
      if (!file.exists())
        throw new XL2Exception("File does not exist" + file.getPath());
      HandleFile = new RandomAccessFile(file,"rw");
      NextHandleLong = HandleFile.length();
      RestoreHandleLong = NextHandleLong;
      file = new File(dbDir,FILE_GARBAGE);
      if (!file.exists())
        throw new XL2Exception("File does not exist" + file.getPath());
      TheGarbageCollector = new GarbageCollector(this,file);
      file = new File(dbDir,FILE_FREE_OBJECT);
      if (!file.exists())
        throw new XL2Exception("File does not exist" + file.getPath());
      ObjFreeList = new FreeList(ObjectWriter.getFile(),file);
      file = new File(dbDir,FILE_FREE_HANDLE);
      if (!file.exists())
        throw new XL2Exception("File does not exist" + file.getPath());
      HandleFreeStack = new FreeStack(file);
      CloseQueue = new ExecutionQueue();
      OpenModeInt = mode;
      TheDaemon = new XL2Daemon(this);
      TheDaemon.start();
      if (TheShutdownHook != null)
        TheShutdownHook.addDatabase(this);
      }
    catch (Exception x) {
      if (ObjectReader != null) {
        try { ObjectReader.close(); } catch (Exception e) {}
        }
      if (ObjectWriter != null) {
        try { ObjectWriter.close(); } catch (Exception e) {}
        }
      if (HandleFile != null) {
        try { HandleFile.close(); } catch (Exception e) {}
        }
      try { HandleFreeStack.close(); } catch (Exception e) {}
      try { ObjFreeList.close(); } catch (Exception e) {}
      try { TheGarbageCollector.close(); } catch (Exception e) {}
      log(null,x);
      if (x instanceof XL2Exception)
        throw (XL2Exception) x;
      throw new XL2Exception(x);
      }
    }

    /**
    * Sets the database root.
    */
  public synchronized void setRoot(Object object) {
    if (OpenModeInt == NOT_OPEN)
      throw new DatabaseNotOpenException(this);
    ((XL2Reference)get(0L)).set(object);
    }


  // Protected Methods
  // -----------------

    /**
    * Evicts all referents that were to be saved during commit.
    */
  protected void abort() {
    try {
      synchronized (HandleFile) {
        HandleFile.setLength(RestoreHandleLong);
        }
      NextHandleLong = RestoreHandleLong;
      TheGarbageCollector.abort();
      synchronized (LockObject) {
        ObjFreeList.abort();
        }
      HandleFreeStack.abort();
      }
    catch (Exception x) {
      throw new XL2Exception(x);
      }
    }

  protected boolean addTransaction(XL2Transaction trans) {
    if (trans == null)
      return false;
    synchronized (XL2TransactionSet) {
      return XL2TransactionSet.add(trans);
      }
    }

    /**
    * Save all changes.
    */
  protected void commit() {
    try {
      RestoreHandleLong = NextHandleLong;
      TheGarbageCollector.commit();
      synchronized (LockObject) {
        ObjFreeList.commit();
        }
      HandleFreeStack.commit();
      try {
        TheGarbageCollector.getFD().sync();
        }
      catch (Exception x) {
        log(null,x);
        }
      try {
        ObjectWriter.getFD().sync();
        }
      catch (Exception x) {
        log(null,x);
        }
      try {
        HandleFreeStack.getFD().sync();
        }
      catch (Exception x) {
        log(null,x);
        }
      }
    catch (XL2Exception x) {
      throw x;
      }
    catch (Exception x) {
      throw new XL2Exception(x);
      }
    }

  protected Object deserialize(InputStream in) 
    throws ClassNotFoundException, IOException {
    return getObjectInput(in).readObject();
    }

    /**
    * Removes referent from handle-referent map.
    */
  protected void evict(Referent referent) {
    TheHandleSet.remove(referent);
    }

    /**
    * Returns the object mapped to the handle.
    */
  protected Object get(long handle) {
    Referent referent = (Referent) TheHandleSet.get(handle);
    Object obj;
    if (referent != null) {
      obj = referent.get();
      if (obj == null)
        ReferentMgr.evict(referent);
      else
        return obj;
      }
    try {
      long offset = 0;
      long refs = 0;
      int len = 0;
      synchronized (HandleFile) {
        HandleFile.seek(handle);
        offset = HandleFile.readLong();
        refs = HandleFile.readLong();
        len = HandleFile.readInt();
        }
      synchronized (LockObject)  {
        ObjectReader.seek(offset + 8L);
        obj = deserialize(ObjectReader);
        }
      referent = new Referent(handle,
        offset,
        refs,
        len,
        obj,
        this,
        true);
      ReferentMgr.add(referent);
      TheHandleSet.add(referent);
      return obj;
      }
    catch (Exception x) {
      throw new XL2Exception(x);
      }
    }

  protected void lock(String dbDir, int mode) 
    throws DatabaseLockedException, IOException {
    File file = new File(dbDir,FILE_WRITE_LOCK);
    if (file.exists())
      throw new DatabaseLockedException(file.getPath());
    if (mode == OPEN_READ_WRITE) {
      file = new File(dbDir);
      String[] readLocks = file.list(
        new FilenameFilter() {
          public boolean accept(File dbDir, String name) {
            if (name.startsWith(FILE_READ_LOCK))
              return true;
            return false;
            }
          });
      if ((readLocks != null) && (readLocks.length > 0))
        throw new DatabaseLockedException(file.getPath());
      //look for read lock files
      LockFile = new File(dbDir,FILE_WRITE_LOCK);
      LockFile.createNewFile();
      LockFile.deleteOnExit();
      }
    else {
      LockFile = new File(dbDir,
                          FILE_READ_LOCK + "." + System.currentTimeMillis());
      LockFile.createNewFile();
      LockFile.deleteOnExit();
      }
    }

    /**
    * Returns the next available handle in the store.
    */
  protected long nextHandle() throws IOException {
    if (HandleFreeStack.isEmpty()) {
      long l = NextHandleLong;
      NextHandleLong += 20;
      return l;
      }
    else 
      return HandleFreeStack.pop();
    }

  protected boolean removeTransaction(XL2Transaction trans) {
    if (trans == null)
      return false;
    synchronized (XL2TransactionSet) {
      return XL2TransactionSet.remove(trans);
      }
    }

  protected void saveObject(Referent referent) throws IOException {
    serialize(referent.get(),Baos);
    int size = Baos.size();
    if (referent.isStored())
      ObjFreeList.add(referent.getOffset(),(referent.getLength() + 8));
    referent.setLength(size);
    referent.setOffset(ObjFreeList.next(size + 8));
    synchronized (LockObject) {
      ObjectWriter.seek(referent.getOffset());
      ObjectWriter.writeLong(referent.ensureHandle());
      ObjectWriter.write(Baos.bytes(),0,size);
      ObjectWriter.flush();
      }
    Baos.reset();
    }

  protected void saveState(Referent referent) throws IOException {
    synchronized (HandleFile) {
      HandleFile.seek(referent.ensureHandle());
      HandleFile.writeLong(referent.getOffset());
      HandleFile.writeLong(referent.getRefCount());
      HandleFile.writeInt(referent.getLength());
      }
    }

  protected void serialize(Object object, OutputStream out) 
    throws IOException {
    ObjectOutput oo = getObjectOutput(out);
    oo.writeObject(object);
    oo.close();
    }

    /**
    * Must never return null.
    */
  protected XL2Transaction[] transactions() {
    synchronized (XL2TransactionSet) {
      return (XL2Transaction[]) 
        XL2TransactionSet.toArray(new XL2Transaction[XL2TransactionSet.size()]);
      }
    }

  protected int transCount() {
    return XL2TransactionSet.size();
    }


  }//XL2Database

/* 
History
-------
v1.1.0, 2000-07-24
  - Added newDatabase.  Subclasses had all the hooks they needed, except the
    ability to create themselves.  Ooops.
  - log replaces closeException.
  - Added calls to RandomAccessFile.getFD().sync()
*/
