// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.odb;

import java.io.*;

  /**
  * Persistent reference.
  * <p><b>Parents and Referents</b><ul>
  * When an XL2Reference is modified (set(Object)), the XL2Reference must be 
  * serialized along with it's encapsulating graph.  The XL2Reference must 
  * know the root object of it's encapsulating graph for transactional locking.
  * This parent (root) is specified during construction of the XL2Reference. 
  * <p>
  * A parent is a first class object.  Objects embedded within the serialized
  * graph of a first class object are second class objects.  XL2References 
  * only point to first class objects and are only constructed with first 
  * class parents. 
  * </ul>
  * <b>Transactional Locking</b><ul>
  * <li>The constructor parameter will be locked.
  * <li>Setting the referent will result in the parent being locked.
  * <li>Setting the referent to a previously unstored object will result
  * in the unstored object being locked as well as the parent.
  * </ul>
  * <b>Memory</b><ul>
  * Referents are not loaded into memory until get() is called.  Referents are
  * softly held so if they have not been transactionally locked and an 
  * application has no strong references to them, they may be 
  * garbage collected.  Referents are not garbage collected until memory is
  * needed by the JVM.
  * </ul>
  * <b>Important</b><br>
  * To clear an XL2Reference, set the referent to null.  For example: <ul>
  * this.XL2ReferenceField.set(null); <br>
  * anotherObject.XL2ReferenceField.set(null); <br>
  * </ul>
  * But never reassign an XL2Reference field.  This will result in a 
  * persistent garbage leak.  The following kinds of statements should 
  * never be written: <ul>
  * this.XL2ReferenceField = null; 
  * anotherObject.XL2ReferenceField = yetAnotherObject.XL2ReferenceField; 
  * </ul>
  * @author Aaron Hansen
  * @see XL2Transaction
  * @since Java 2
  * @version 1.0.0
  */
public class XL2Reference implements Externalizable {

  // Protected Fields
  // ----------------

  protected boolean ChangedBoolean = false;

    /**
    * Current value of the ref.
    */
  protected Object CurrentObject = null;

    /**
    * Placement within a store.
    */
  protected long HandleLong = -1;

    /**
    * Value of ref when first fetched in a runtime.
    */
  protected Object OriginalObject = null;

    /**
    * Who to add to the current transaction whenever the referent is changed.
    */
  protected Object ParentObject = null;


  // Constructors
  // ------------

  public XL2Reference() {}

    /**
    * Constructs an XL2Reference with itself as the parent such that
    * changing the ref will lock itself with the current XL2Transaction.  
    * The parameter has no meaning.
    * @param meaningless Has no meaning.
    * @see #XL2Reference(Object)
    */
  public XL2Reference(int meaningless) {
    ParentObject = this;
    }

    /**
    * Constructs an XL2Reference with the specified parent.  Must be done
    * in a transaction.
    * @throws XL2Exception When the calling thread is not joined
    *         to a transaction.
    * @throws InvalidObjectException When the parent was previously locked by a
    *         transaction that was aborted.
    * @throws LockNotGrantedException When the parent has a conflicting lock 
    *         with another transaction.
    * @see XL2Transaction
    */
  public XL2Reference(Object parent) 
    throws InvalidObjectException, LockNotGrantedException {
    XL2Transaction trans = XL2TransactionMgr.currentTransaction();
    if (trans == null)
      throw new XL2Exception("No transaction in progress.");
    else
      trans.writeLock(parent);
    ParentObject = parent;
    }


  // Public Methods
  // --------------

    /**
    * Returns the referent.  Fetches the object from the database if needed.
    * @throws XL2Exception If no transaction is in progress.
    */
  public Object get() {
    if (CurrentObject == null) {
      if (HandleLong == -1L)
        return null;
      if (!ChangedBoolean) {
        XL2Transaction trans = XL2TransactionMgr.currentTransaction();
        if (trans == null)
          throw new XL2Exception("No transaction in progress.");
        OriginalObject = trans.getDatabase().get(HandleLong);
        CurrentObject = OriginalObject;
        }
      }
    return CurrentObject;
    }

    /**
    * For serialization, not for applications.
    * @see java.io.Externalizable
    */
  public void readExternal(ObjectInput in) 
    throws ClassNotFoundException, IOException {
    HandleLong = in.readLong();
    ParentObject = in.readObject();
    }

    /**
    * Changes the referent.
    * If a parent was specified during construction, then it will be locked by 
    * the current XL2Transaction.  If the parameter is not currently stored 
    * in the database, it will be locked by the current transaction.
    * @throws XL2Exception When the current thread is not
    *         joined to a transaction.
    * @throws InvalidObjectException If the object was previously locked by a 
    *         transaction that was aborted.
    * @throws LockNotGrantedException When either the parent or referent is 
    *         locked by another transaction.
    * @param obj The new referent.
    */
  public void set(Object obj) 
    throws InvalidObjectException, LockNotGrantedException {
    XL2Transaction trans = XL2TransactionMgr.currentTransaction();
    if (trans == null)
      throw new XL2Exception("No transaction in progress.");
    if (ParentObject != null)
      trans.writeLock(ParentObject);
    if ((OriginalObject == null) && (HandleLong >= 0) && !ChangedBoolean)
      get();
    CurrentObject = obj;
    Referent referent = ReferentMgr.get(obj);
    if (referent == null) 
      trans.writeLock(CurrentObject);
    ChangedBoolean = true;
    }

    /**
    * For serialization, not for applications.
    * @see java.io.Externalizable
    */
  public void writeExternal(ObjectOutput out) throws IOException {
    XL2Transaction trans = XL2TransactionMgr.currentTransaction();
    if (ChangedBoolean) {
      XL2Database db = trans.getDatabase();
      Referent referent;
      Object object;
      if ((OriginalObject == null) && (HandleLong != -1L))
        OriginalObject = db.get(HandleLong);
      if ((OriginalObject != null) && (OriginalObject != CurrentObject)) {
        referent = ReferentMgr.get(OriginalObject);
        referent.decRefCount();
        }
      if (CurrentObject == null) {
        HandleLong = -1L;
        }
      else {
        referent = ReferentMgr.get(CurrentObject);
        HandleLong = referent.ensureHandle();
        referent.incRefCount();
        }
      OriginalObject = CurrentObject;
      }
    out.writeLong(HandleLong);
    out.writeObject(ParentObject);
    ChangedBoolean = false;
    }


  }//XL2Reference
