// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.odb;

import java.util.Iterator;
import xl2.odb.util.HashSet;
import xl2.odb.util.ReferenceSet;
import xl2.transaction.AbstractTransaction;
import xl2.transaction.AbortFailedException;
import xl2.transaction.TransactionAbortedException;
import xl2.transaction.TransactionNotClosedException;
import xl2.transaction.TransactionNotOpenException;

  /**
  * A transaction of a single XL2database.
  * <p>
  * Like thread cooperation using the <i>synchronized</i> keyword, 
  * applications must cooperate by successfully locking objects before 
  * using them.  Any transaction can modify another XL2Transaction's object if
  * no attempt is made to lock the object first!
  * <p><b>Locking</b><br>
  * New and modified objects get saved to the database via locking.  If
  * an object is not locked, it may be lost.
  * XL2References automatically lock objects, but for modifications to second 
  * class objects and primitive fields, one must manually lock parent first
  * class objects.
  * <ul>
  * <li>Many transactions may read lock the same object.
  * <li>Only one transaction may write lock an object.
  * <li>An object can not be read locked and write locked at the same time.
  * <li>A transaction may upgrade a read lock to a write lock, provided
  * it is the only transaction holding a read lock.  A write lock may not
  * be downgraded to a read lock.
  * <li>Write locked objects are serialized into the database during commital.
  * <li>Read only transactions may only read lock objects.
  * <li>Objects may not be unlocked, the transaction must be aborted.
  * <li>InvalidObjectExceptions are thrown when transactions try to lock 
  * in-memory instances that may be inconsistent with their last committed 
  * state. For example, objects that were previously locked by aborted 
  * transactions.
  * </ul>
  * <p><b>Threads</b><br>
  * Threads must be joined to transactions.  This is so XL2References
  * can "find" their database.
  * <ul>
  * <li>A thread can only be joined to one transaction at a time.
  * <li>A transaction may have many threads joined to it at once.
  * <li>If a transaction is set as the global transaction, any thread
  * not joined to a transaction is automatically joined to the global
  * transaction.
  * </ul>
  * @author Aaron Hansen
  * @see InvalidObjectException
  * @see LockNotGrantedException
  * @see XL2TransactionMgr#currentTransaction()
  * @since Java 2
  * @version 1.0.0
  */
public class XL2Transaction extends AbstractTransaction {

  // Public Fields
  // -------------

    /**
    * Lock type.
    */
  public static int READ = XL2Constants.READ;

    /**
    * Lock type.
    */
  public static int WRITE = XL2Constants.WRITE;


  // Protected Fields
  // ----------------

    /**
    * The System.currentTimeMillis at which the transaction will be 
    * automatically aborted. 0 or less is an infinite lifetime.
    */
  protected long ExpiresLong = 0L;

    /**
    * The objects whose state needs to be stored.
    */
  protected HashSet HandleSet = null;

    /**
    * If true, object can not be write locked.
    */
  protected boolean ReadOnlyBoolean = false;

    /**
    * Objects to preserve.
    */
  protected HashSet ReadSet = null;

    /**
    * XL2Database of the transaction.
    */
  protected XL2Database TheXL2Database = null;
  
    /**
    * The length of time after which an open transaction will be automatically
    * aborted.
    */
  protected long TimeoutLong = 0L;

    /**
    * Objects to store.
    */
  protected HashSet WriteSet = null;


  // Constructors
  // ------------

    /**
    * Constructs a transaction for a single database.
    * The thread of construction is automatically joined to the transaction.
    * @param db XL2Database of the transaction. Can not be null.
    * @throws DatabaseNotOpenException
    * @throws XL2Exception If anything goes wrong.
    * @see #join()
    */
  protected XL2Transaction(XL2Database db) {
    if (db == null)
      throw new XL2Exception("Null database.");
    if (db.getOpenMode() == db.NOT_OPEN)
      throw new DatabaseNotOpenException(db);
    TheXL2Database = db;
    join();
    }


  // Public Methods
  // --------------

  public void abort() throws 
    //Here for doc clarity.
    AbortFailedException,TransactionNotOpenException {
    super.abort();
    }

  public void begin() throws TransactionNotClosedException {
    //Here for doc clarity.
    super.begin();
    }

  public void commit() throws 
    //Here for doc clarity.
    TransactionAbortedException, TransactionNotOpenException {
    super.commit();
    }

    /**
    * The database of the transaction.
    */
  public XL2Database getDatabase() {
    return TheXL2Database;
    }

    /**
    * The time (System.currentTimeMillis) after which the open transaction 
    * will be automatically aborted.  This only has meaning when the transaction
    * is open.
    * @return &lt;=0 Is an infinite expiration, and the default.
    */
  public long getExpiration() {
    return ExpiresLong;
    }

    /**
    * The length of time after which an open transaction will be automatically
    * aborted.
    * @return &lt;=0 Is an infinite lifetime, and the default.
    */
  public long getTimeout() {
    return TimeoutLong;
    }

    /**
    * Returns true if the object is read locked by one or more transactions.
    */
  public static boolean isReadLocked(Object obj) {
    Referent referent = ReferentMgr.get(obj);
    if (referent == null)
      return false;
    return (LockMgr.getLockMode(obj,referent.getHandle()) == READ);
    }

    /**
    * True if the transaction can not write lock objects.
    */
  public boolean isReadOnly() {
    return ReadOnlyBoolean;
    }

    /**
    * Returns true if the object is write locked by any transaction.
    */
  public static boolean isWriteLocked(Object obj) {
    Referent referent = ReferentMgr.get(obj);
    if (referent == null)
      return false;
    return (LockMgr.getLockMode(obj,referent.getHandle()) == WRITE);
    }

    /**
    * Attempts to lock the object with the given mode.
    * A READ lock prevents another transaction from obtaining a WRITE lock.  A
    * WRITE lock prevents another transaction from obtaining any lock, and
    * a WRITE locked object will be serialized into the database upon committal.
    * @throws InvalidObjectException
    * @throws TransactionNotOpenException
    * @throws LockNotGrantedException
    */
  public void lock(Object object, int mode) throws InvalidObjectException, 
                                                   LockNotGrantedException, 
                                                   TransactionNotOpenException {
    if (object == null)
      return; 
    if (!this.isOpen()) {
      throw new TransactionNotOpenException(
        this, ("Can't lock: " + String.valueOf(object)));
      }
    if ((mode == WRITE) && ReadOnlyBoolean) {
      throw new LockNotGrantedException(
        object,LockNotGrantedException.READONLY_TRANSACTION);
      }
    synchronized (XL2Transaction.class) {
      Referent referent = ReferentMgr.get(object);
      if (InvalidMgr.contains(object))
        throw new InvalidObjectException(object);
      if (referent == null) {
        referent = 
          new Referent( -1L, -1L, 0L, -1, object, TheXL2Database, false);
        ReferentMgr.add(referent);
        }
      if (mode == READ) {
        if (ReadSet.contains(object) || WriteSet.contains(object))
          return;
        LockMgr.lock(object,referent.getHandle(),READ);
        ReadSet.add(object);
        }
      else if (mode == WRITE) {
        if (WriteSet.contains(object))
          return;
        if (ReadSet.remove(object) || HandleSet.contains(object))
          LockMgr.upgrade(object,referent.getHandle());
        else
          LockMgr.lock(object,referent.getHandle(),WRITE);
        WriteSet.add(object);
        HandleSet.add(object);
        }
      }
    }

    /**
    * A convenience function that calls lock(object,XL2Transaction.READ).
    * @see #lock(Object,int)
    */ 
  public void readLock(Object object) {
    lock(object,READ);
    }

    /**
    * Only to be called when the transaction is closed.  When true, 
    * prevents the transaction write locking objects.
    * @throws XL2TransactionNotClosedException
    */
  public synchronized void setReadOnly(boolean readonly) {
    if (this.isOpen())
      throw new TransactionNotClosedException(this);
    ReadOnlyBoolean = readonly;
    }

    /**
    * The length of time after which an open transaction will be automatically
    * aborted. If the transaction is closed when this is called, the 
    * lifetime begins when the transaction begins. If the transaction 
    * is open when this is called, the lifetime is reset, beginning
    * at the time of the method call.
    * @param millis &lt;=0 Is an infinite lifetime, and the default.
    */
  public synchronized void setTimeout(long millis) {
    TimeoutLong = millis;
    if (millis > 0)
      ExpiresLong = System.currentTimeMillis() + millis;
    else
      ExpiresLong = 0;
    }

    /**
    * Same as lock(Object,int) except no exception is thrown.
    * @return False if the lock is not granted.
    */
  public boolean tryLock(Object obj, int mode) {
    try {
      lock(obj,mode);
      return true;
      }
    catch (Exception x) {
      return false;
      }
    }

    /**
    * A convenience function that calls lock(object,XL2Transaction.WRITE).
    * @see #lock(Object,int)
    */ 
  public void writeLock(Object object) {
    lock(object,WRITE);
    }


  // Protected Methods
  // -----------------

  protected long checkExpiration() {
    if ( (ExpiresLong > 0) 
      && (ExpiresLong <= System.currentTimeMillis())
      && (this.isOpen()) )
      this.abort();
    if (this.isOpen())
      return ExpiresLong;
    else
      return 0L;
    }

  protected void doAbort() {
    TheXL2Database.CloseQueue.enqueue();
    XL2Transaction trans = XL2TransactionMgr.currentTransaction();
    if ((trans != this) && (trans != null))
      join();
    if (HandleSet != null) {
      Iterator iterator = HandleSet.iterator();
      Referent referent;
      Object object;
      while (iterator.hasNext()) {
        object = iterator.next();
        referent = ReferentMgr.get(object);
        if (referent.isStored())
          InvalidMgr.add(object);
        LockMgr.unlock(object,referent.getHandle());
        ReferentMgr.evict(referent);
        }
      }
    doClose();
    if ((trans != this) && (trans != null) && (trans.isOpen()))
      trans.join();
    TheXL2Database.CloseQueue.dequeue();
    }

  protected void doBegin() {
    join();
    if (TimeoutLong > 0) {
      ExpiresLong = System.currentTimeMillis() + TimeoutLong;
      TheXL2Database.TheDaemon.hello();
      }
    else
      ExpiresLong = 0;
    if (TheXL2Database.getOpenMode() == XL2Database.OPEN_READ_ONLY) {
      ReadOnlyBoolean = true;
      }
    else {
      if (WriteSet == null)
        WriteSet = new HashSet();
      if (HandleSet == null)
        HandleSet = new HashSet();
      }
    ReadSet = new HashSet();
    TheXL2Database.addTransaction(this);
    }

    /**
    * Called by abort and commit.
    */
  protected void doClose() {
    if (ReadSet != null) { //remove read locks
      Iterator iterator = ReadSet.iterator();
      Referent ref;
      Object obj;
      while (iterator.hasNext()) {
        obj = iterator.next();
        ref = ReferentMgr.get(obj);
        LockMgr.unlock(obj,ref.getHandle());
        }
      ReadSet.clear();
      }
    if (WriteSet != null)
      WriteSet.clear();
    if (HandleSet != null)
      HandleSet.clear();
    TheXL2Database.removeTransaction(this);
    }

  protected void doCommit() {
    TheXL2Database.CloseQueue.enqueue();
    XL2Transaction trans = XL2TransactionMgr.currentTransaction();
    try {
      if ((trans != this) && (trans != null))
        join();
      boolean loopAgain = true;
      Iterator iterator;
      Referent referent;
      if (WriteSet != null) {
        while (loopAgain && (WriteSet.size() > 0)) {
          loopAgain = false;
          iterator = WriteSet.iterator();
          while (iterator.hasNext()) {
            referent = ReferentMgr.get(iterator.next());
            if (referent.getRefCount() > 0) {
              TheXL2Database.saveObject(referent);
              referent.setStored(true);
              loopAgain = true;
              iterator.remove();
              }
            }
          }
        }
      if (HandleSet != null) {
        Object obj;
        iterator = HandleSet.iterator();
        while (iterator.hasNext()) {
          obj = iterator.next();
          referent = ReferentMgr.get(obj);
          if (referent.getRefCount() == 0) {
            if (referent.isStored()) {
              TheXL2Database.TheGarbageCollector.add(referent);
              TheXL2Database.saveState(referent);
              }
            else 
              ReferentMgr.evict(referent);
            }
          else 
            TheXL2Database.saveState(referent);
          LockMgr.unlock(obj,referent.getHandle());
          }
        }
      TheXL2Database.commit();
      doClose();
      if ((trans != this) && (trans != null) && (trans.isOpen()))
        trans.join();
      TheXL2Database.CloseQueue.dequeue();
      }
    catch (Exception x) {
      try {
        TheXL2Database.log(null,x);
        abort(); 
        TheXL2Database.abort();
        }
      catch (Exception e) {
        TheXL2Database.log("Failure aborting transaction.",e);
        }
      if ((trans != this) && (trans != null) && (trans.isOpen()))
        trans.join();
      throw new TransactionAbortedException(this,x);
      }
    }

  protected void saveState(Object object) {
    if (!HandleSet.contains(object)) {
      Referent ref = ReferentMgr.get(object);
      LockMgr.lock(object,ref.getHandle(),READ);
      HandleSet.add(object);
      }
    }


  }//Transaction
