// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.transaction;

import java.util.*;

  /**
  * Skeletal implementation of the Transaction interface.  Subclasses need
  * only over-ride the following protected methods: 
  * <ul>
  * <li>protected void doAbort()
  * <li>protected void doBegin()
  * <li>protected void doCommit()
  * </ul>
  * @author Aaron Hansen
  * @version 1.0.0
  */
public class AbstractTransaction implements Transaction {

  // Protected Fields
  // ----------------

  protected int StatusInt = STATUS_INSTANTIATED;


  // Public Methods
  // --------------

    // See the interface documentation for this method.
  public synchronized void abort() 
    throws AbortFailedException, TransactionNotOpenException{
    if (isClosed())
      throw new TransactionNotOpenException(this);
    try {
      StatusInt = STATUS_ABORTING;
      doAbort();
      }
    catch (Exception x) {
      StatusInt = STATUS_ABORT_FAILED;
      throw new AbortFailedException(this,x);
      }
    StatusInt = STATUS_ABORTED;
    }

    // See the interface documentation for this method.
  public synchronized void begin() throws TransactionNotClosedException {
    if (!isClosed())
      throw new TransactionNotClosedException(this);
    try {
      StatusInt = STATUS_BEGINNING;
      doBegin();
      StatusInt = STATUS_IN_PROGRESS;
      }
    catch (TransactionAbortedException x) {  //doBegin may call abort
      throw x;
      }
    catch (Exception x) {
      if (StatusInt == STATUS_BEGINNING)
        abort();
      throw new TransactionAbortedException(this,x);
      }
    }

    // See the interface documentation for this method.
  public synchronized void commit()
    throws TransactionAbortedException, TransactionNotOpenException {
    if (!isOpen())
      throw new TransactionNotOpenException(this);
    else if (StatusInt == STATUS_MARKED_ABORT) {
      abort();
      throw new TransactionAbortedException(this,
        "Transaction.setAbortOnly() was called.");
      }
    try {
      StatusInt = STATUS_COMMITTING;
      doCommit();
      StatusInt = STATUS_COMMITTED;
      }
    catch (TransactionAbortedException x) {  //doCommit may call abort
      throw x;
      }
    catch (Exception x) {
      if (StatusInt == STATUS_COMMITTING)
        abort();
      throw new TransactionAbortedException(this,x);
      }
    }

    // See the interface documentation for this method.
  public int getStatus() {
    return StatusInt;
    }

    // See the interface documentation for this method.
  public boolean isClosed() {
    int status = StatusInt;
    return ( (status == STATUS_ABORTED)
          || (status == STATUS_COMMITTED)
          || (status == STATUS_INSTANTIATED) );
    }

    // See the interface documentation for this method.
  public boolean isOpen() {
    int status = StatusInt;
    return ( (status == STATUS_IN_PROGRESS) 
          || (status == STATUS_MARKED_ABORT));
    }

    // See the interface documentation for this method.
  public void join() {
    TransactionMgr.join(this);
    }

    // See the interface documentation for this method.
  public void leave() {
    TransactionMgr.leave(this);
    }

    // See the interface documentation for this method.
  public void setAbortOnly() throws TransactionNotOpenException {
    if (!isOpen())
      throw new TransactionNotOpenException(this);
    StatusInt = STATUS_MARKED_ABORT;
    }


  // Protected Methods
  // -----------------

    /**
    * Abort has been called and the status of the transaction is 
    * STATUS_ABORTING. Avoid throwing an exception from this method.
    * The status during this method will be STATUS_ABORTING.
    */
  protected void doAbort() {}

    /**
    * Begin has been called and the status of the transaction is 
    * STATUS_BEGINNING.  Upon the completion of begin the 
    * status of the transaction will be STATUS_IN_PROGRESS.
    * Throw an exception to abort the transaction.
    */
  protected void doBegin() {}

    /**
    * Commit has been called and the status of the transaction is 
    * STATUS_COMMITTING.  Do not call abort or begin.  
    * Throw an exception to abort the transaction.
    */
  protected void doCommit() {}


  }//Transaction
