// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.transaction;

  /**
  * Basic interface that defines the state boundries for a set of 
  * resoures.
  * @author Aaron Hansen
  * @version 1.0.0
  */
public interface Transaction {

  // Constants
  // ---------

    /**
    * Transaction aborted - closed state.
    * @see #getStatus()
    */
  public static int STATUS_ABORTED = 0;

    /**
    * Transaction aborted - closed state.
    * @see #getStatus()
    */
  public static int STATUS_ABORT_FAILED = 1;

    /**
    * Transaction in the process of aborting.
    * @see #getStatus()
    */
  public static int STATUS_ABORTING = 2;

    /**
    * Transaction in the process of beginning.
    * @see #getStatus()
    */
  public static int STATUS_BEGINNING = 3;

    /**
    * Transaction has been committed - closed state.
    * @see #getStatus()
    */
  public static int STATUS_COMMITTED = 4;

    /**
    * Transaction is committing.
    * @see #getStatus()
    */
  public static int STATUS_COMMITTING = 5;

    /**
    * Transaction is in progress and has not been marked STATUS_MARKED_ABORT
    * - open state.
    * @see #getStatus()
    */
  public static int STATUS_IN_PROGRESS = 6;

    /**
    * Transaction has been constructed and not yet begun - closed state.
    * @see #getStatus()
    */
  public static int STATUS_INSTANTIATED = 7;

    /**
    * Transaction is in progress, and can only abort - open state.
    * @see #getStatus()
    */
  public static int STATUS_MARKED_ABORT = 8;


  // Public Methods
  // --------------

    /**
    * Rollback the transaction.  If an exception occurs while aborting the
    * status of the transaction becomes STATUS_ABORT_FAILED.  
    * Only abort can be called if the status is STATUS_ABORT_FAILED.
    * @throws AbortFailedException
    * @throws TransactionNotOpenException
    */
  public void abort() throws AbortFailedException, TransactionNotOpenException;

    /**
    * Calling begin multiple times on the same transaction, without 
    * an intervening call to commit or abort, causes 
    * TransactionNotClosedException to be thrown on the second 
    * and subsequent calls.  
    * @throws TransactionNotClosedException
    */
  public void begin() throws TransactionNotClosedException;

    /**
    * Save the transaction state.
    * @throws AbortedFailedException If the transaction attempts to abort and
    *                                an exception occurrs.
    * @throws TransactionAbortedException If an exeception occurs.
    * @throws TransactionNotOpenException
    * @see #isOpen()
    */
  public void commit() 
    throws TransactionAbortedException, TransactionNotOpenException;

    /**
    * See Transaction STATUS_??? fields.
    */
  public int getStatus();

    /**
    * True if the status aborted, committed, or instantiated.
    */
  public boolean isClosed();

    /**
    * True if the status is in progress or marked for abort.
    */
  public boolean isOpen();
  
    /**
    * Associates the calling thread to the transaction.  A thread can be
    * joined to at most one transaction at any given time.
    * @see #leave()
    */
  public void join();

    /**
    * Disassociates the calling thread from the transaction. 
    * @see #join()
    */
  public void leave();

    /**
    * Modifies the transaction such that the only possible outcome is to 
    * abort the transaction.  This has no effect on the state of the 
    * transaction until commit() is called.
    * @throws TransactionNotOpenException
    */
  public void setAbortOnly() throws TransactionNotOpenException;


  }//Transaction
