// Copyright (c) 2000 by Aaron Hansen. Use is subject to license terms.

package xl2.transaction;

import java.util.*;

  /**
  * Provides static access to threads' transactions.
  * @author Aaron Hansen
  * @version 1.0.0
  */
public class TransactionMgr {

  // Protected Fields
  // ----------------

  protected static Transaction GlobalTransaction = null;

    /**
    * Threads and their associate transactions.
    */
  protected static Map ThreadMap = 
    Collections.synchronizedMap(new WeakHashMap());


  // Public Methods
  // --------------

    /**
    * Calls abort on the current transaction.
    * @throws TransactionRequiredException If the calling thread is not joined
    *         to a transaction and there is no global transaction.
    * @see Transaction#abort()
    */
  public static void abort() {
    Transaction t = currentTransaction();
    if (t == null)
      throw new TransactionRequiredException();
    t.abort();
    }

    /**
    * Calls begin on the current transaction.
    * @throws TransactionRequiredException If the calling thread is not joined
    *         to a transaction and there is no global transaction.
    * @see Transaction#begin()
    */
  public static void begin() {
    Transaction t = currentTransaction();
    if (t == null)
      throw new TransactionRequiredException();
    t.begin();
    }

    /**
    * Calls commit on the current transaction.
    * @throws TransactionRequiredException If the calling thread is not joined
    *         to a transaction and there is no global transaction.
    * @see Transaction#commit()
    */
  public static void commit() {
    Transaction t = currentTransaction();
    if (t == null)
      throw new TransactionRequiredException();
    t.commit();
    }

    /**
    * The transaction of the calling thread.  If the calling thread
    * has not joined any transactions, the global transaction is returned.  
    * @return Possibly null.
    * @see #getGlobal()
    */
  public static Transaction currentTransaction() {
    Transaction trans = (Transaction) ThreadMap.get(Thread.currentThread());
    if (trans == null)
      return getGlobal();
    return trans;
    }

    /**
    * The default transaction for any thread not explicitly joined to one.
    * @return Possibly null.
    * @see #setGlobal(Transaction)
    */
  public static Transaction getGlobal() {
    return GlobalTransaction;
    }

    /**
    * If currentTransaction() can not find a transaction for the calling
    * thread, it will return the transaction specified here.
    * @param trans The default transaction, if there is one.
    */
  public static void setGlobal(Transaction trans) {
    GlobalTransaction = trans;
    }


  // Protected Methods
  // -----------------

    /**
    * Associates the calling thread to the transaction.  A thread can be
    * joined to at most one transaction at any given time.
    */
  protected static void join(Transaction trans) {
    ThreadMap.put(Thread.currentThread(),trans);
    }

    /**
    * Disassociates the calling thread from the transaction. 
    */
  protected static void leave(Transaction trans) {
    Thread t = Thread.currentThread();
    if (ThreadMap.get(t) == trans)
      ThreadMap.remove(t);
    }


  }//TransactionMgr
