package xl2.util;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Vector;

  /**
  * A cron-like schedule. <br>
  * <b>Possible values</b>
  * <ul>
  * <li>Minute, -1 for any minute, or 0-59
  * <li>Hour Of Day, -1 for any hour of the day, or 0-23
  * <li>Day Of Week, -1 for any day of the week, or 1-7
  * <li>Day Of Month, -1 for any day of the month, or 1-31
  * <li>Month, -1 for any month of the year, or 1-12
  * </ul>
  * <b>Constraints</b>
  * <ul>
  * <li>If the day of week is -1 and the day of month is 1-31, then
  * the time is only valid on the day of month.
  * <li>If the day of week is 1-7 and the day of month is -1, then
  * the time is valid on the day of week.
  * <li>If both the day of week and and day of month are positive, then
  * time is valid on days when both values match.
  * <li>If the day of week, day of month, and month are positive, then all
  * three values have to be valid.
  * </ul>
  * @author aaron_hansen@yahoo.com
  */
public class Schedule {

  //public methods
  //--------------

  protected int MinuteInt = -1;

  protected int HourInt = -1;

  protected int DayWeekInt = -1;

  protected int DayMonthInt = -1;

  protected int MonthInt = -1;


  //constructors
  //------------

  public Schedule() {}

    /**
    * @param min -1 for any min, or 0-59
    * @param hour -1 for any hour, or 0-23
    * @param dayWk -1 for any day of the week, or 1-7
    * @param dayMo -1 for any day of the month, or 1-31
    * @param month -1 for any month, or 1-12
    * @see #init(int,int,int,int,int)
    */
  public Schedule(int min,
                  int hour,
                  int dayWk,
                  int dayMo,
                  int month) {
    init(min,hour,dayWk,dayMo,month);
    }

    /**
    * Parses the schedule from the string.  The schedule may be null
    * indicating run anytime.  Otherwise the schedule must contain
    * minute, hour, dayWk, dayMo, and month seperated by whitespace.
    * @param schedule Minute, hour, dayWk, dayMo, and month seperated by
    *                 whitespace.  
    * @see #init(int,int,int,int,int)
    */
  public Schedule(String schedule) {
    if (schedule == null)
      return;
    int min = -1;
    int hour = -1;
    int dayWk = -1;
    int dayMo = -1;
    int month = -1;
    Vector times = Strings.fields(schedule,' ');
    if (times.size() != 5)
      throw new IllegalArgumentException("Invalid schedule: " + schedule);
    String temp = null;
    try {
      temp = times.elementAt(0).toString();
      min = Integer.parseInt(temp);
      }
    catch (NumberFormatException e) {
      throw new IllegalArgumentException("Invalid minute: " + temp);
      }
    try {
      temp = times.elementAt(1).toString();
      hour = Integer.parseInt(temp);
      }
    catch (NumberFormatException e) {
      throw new IllegalArgumentException("Invalid hour: " + temp);
      }
    try {
      temp = times.elementAt(2).toString();
      dayWk = Integer.parseInt(temp);
      }
    catch (NumberFormatException e) {
      throw new IllegalArgumentException("Invalid day of week: " + temp);
      }
    try {
      temp = times.elementAt(3).toString();
      dayMo = Integer.parseInt(temp);
      }
    catch (NumberFormatException e) {
      throw new IllegalArgumentException("Invalid day of month: " + temp);
      }
    try {
      temp = times.elementAt(4).toString();
      month = Integer.parseInt(temp);
      }
    catch (NumberFormatException e) {
      throw new IllegalArgumentException("Invalid month: " + temp);
      }
    init(min,hour,dayWk,dayMo,month);
    }


  //public methods
  //--------------

    /**
    * Initializes the state of the Schedule.
    * @param min -1 for any min, or 0-59
    * @param hour -1 for any hour, or 0-23
    * @param dayWk -1 for any day of the week, or 1-7
    * @param dayMo -1 for any day of the month, or 1-31
    * @param month -1 for any month, or 1-12
    */
  public void init(int min,
                   int hour,
                   int dayWk,
                   int dayMo,
                   int month) {
    if ((min != -1) && (min < 0) || (dayWk > 23))
      throw new IllegalArgumentException("Minute must be -1 or 0-23");
    if ((hour != -1) && (hour < 0) || (dayWk > 23))
      throw new IllegalArgumentException("Hour must be -1 or 0-23");
    if ((dayWk != -1) && (dayWk < 1) || (dayWk > 7))
      throw new IllegalArgumentException("Day of week must be -1 or 1-7");
    if ((dayMo != -1) && (dayMo < 1) || (dayMo > 31))
      throw new IllegalArgumentException("Day of month must be -1 or 1-31");
    if ((month != -1) && (month < 1) || (month > 12))
      throw new IllegalArgumentException("Month must be -1 or 1-12");
    MinuteInt = min;
    HourInt = hour;
    DayWeekInt = dayWk;
    DayMonthInt = dayMo;
    MonthInt = month;
    }

    /**
    * Returns true if the calendar matches the schedule.
    */
  public boolean isValid(Calendar calendar) {
    return isValid(MinuteInt,HourInt,DayWeekInt,DayMonthInt,MonthInt,calendar);
    }

    /**
    * Returns true if the calendar matches the schedule.
    * @param min -1 for any min, or 0-59
    * @param hour -1 for any hour, or 0-23
    * @param dayWk -1 for any day of the week, or 1-7
    * @param dayMo -1 for any day of the month, or 1-31
    * @param month -1 for any month, or 1-12
    * @param calendar A point in time for comparing against the schedule.
    */
  public static boolean isValid(int min,
                                int hour,
                                int dayWk,
                                int dayMo,
                                int month,
                                Calendar calendar) {
    if ((min > 0) && (calendar.get(calendar.MINUTE) != min))
      return false;
    if ((hour > 0) && (calendar.get(calendar.HOUR_OF_DAY) != hour))
      return false;
    if ((dayWk > 0) && (calendar.get(calendar.DAY_OF_WEEK) != dayWk))
      return false;
    if ((dayMo > 0) && (calendar.get(calendar.DAY_OF_MONTH) != dayMo))
      return false;
    if ((month >= 0) && (calendar.get(calendar.MONTH) != month))
      return false;
    return true;
    }

    /**
    * Calculates the next run time given the schedule.
    * @param reference Reference (time in ms) for determining next run.
    * @return The time in milliseconds of the next valid time.
    * @throws IllegalStateException If the next run time is beyond
    *                               five calendar years.
    */
  public long calculate(long reference) {
    return calculate(MinuteInt,
                     HourInt,
                     DayWeekInt,
                     DayMonthInt,
                     MonthInt,
                     reference);
    }

    /**
    * Calculates the next valid time given the cron-like schedule.  Will only
    * look forward 5 calendar years before throwing an IllegalStateException.
    * @param min -1 for any min, or 0-59
    * @param hour -1 for any hour, or 0-23
    * @param dayWk -1 for any day of the week, or 1-7
    * @param dayMo -1 for any day of the month, or 1-31
    * @param month -1 for any month, or 1-12
    * @param reference Time in ms for calculate next valid time.
    * @return The time in milliseconds of the next scheduled event.
    * @throws IllegalStateException If the next scheduled run time is beyond
    *                               five calendar years.
    */
  public static long calculate(int min,
                               int hour,
                               int dayWk,
                               int dayMo,
                               int month,
                               long reference) {
    if ((min != -1) && (min < 0) || (dayWk > 23))
      throw new IllegalArgumentException("Minute must be -1 or 0-23");
    if ((hour != -1) && (hour < 0) || (dayWk > 23))
      throw new IllegalArgumentException("Hour must be -1 or 0-23");
    if ((dayWk != -1) && (dayWk < 1) || (dayWk > 7))
      throw new IllegalArgumentException("Day of week must be -1 or 1-7");
    if ((dayMo != -1) && (dayMo < 1) || (dayMo > 31))
      throw new IllegalArgumentException("Day of month must be -1 or 1-31");
    if ((month != -1) && (month < 1) || (month > 12))
      throw new IllegalArgumentException("Month must be -1 or 1-12");
    if (month > 0) //Calendar actually uses 0-11
      month--;
    Calendar c = new GregorianCalendar();
    c.setTime(new Date(reference));
    if (min != -1)
      c.set(c.MINUTE,min);
    if (hour != -1)
      c.set(c.HOUR_OF_DAY,hour);
    long time = c.getTime().getTime();
    if ((time > reference) && isValid(min, hour, dayWk, dayMo, month,c))
      return time;
    int temp;
    if (min == -1) {
      temp = c.MINUTE;
      while (c.get(temp) < 59) {
        c.roll(temp,true);
        if (isValid(min,hour,dayWk,dayMo,month,c) 
          && ((time = c.getTime().getTime()) > reference))
          return time;
        }
      c.set(temp,0);
      }
    if (hour == -1) {
      temp = c.HOUR_OF_DAY;
      while (c.get(temp) < 23) {
        c.roll(temp,true);
        if (isValid(min,hour,dayWk,dayMo,month,c) 
          && ((time = c.getTime().getTime()) > reference))
          return time;
        }
      c.set(temp,0);
      }
    temp = c.DAY_OF_YEAR;
    int maxDoy;//max day of year
    int chod = c.HOUR_OF_DAY; //jdk1.2.2 Calendar bug workaround
    for (int i = 5; --i >= 0;) {
      maxDoy = c.getActualMaximum(temp);
      while (c.get(temp) < maxDoy) {
        c.roll(temp,true);
        //As of jdk1.2.2, Calendar.roll does not handle day light savings very 
        //well. Rolling DAY_OF_YEAR on that day rolls hour 0 to hour 23 of the 
        //same day!
        c.set(chod,(hour > 0 ? hour : 0)); //jdk1.2.2 bug workaround
        if (isValid(min,hour,dayWk,dayMo,month,c) 
          && ((time = c.getTime().getTime()) > reference))
          return time;
        }
      c.roll(temp,true);//would otherwise roll to the last day of the next year
      c.roll(c.YEAR,true);
      }
    throw new IllegalStateException("Next time too far into the future.");
    }

  public String toString() {
    StringBuffer buff = new StringBuffer();
    buff.append("Min=" + MinuteInt);
    buff.append(":Hour=" + HourInt);
    buff.append(":DayWeek=" + DayWeekInt);
    buff.append(":DayMonth=" + DayMonthInt);
    buff.append(":Month=" + (MonthInt + 1));
    return buff.toString();
    }

  }//class Job
